/*******************************************************************************
    Copyright (c) 2003-2016 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/


#ifndef __gh100_dev_fault_h__
#define __gh100_dev_fault_h__
/* This file is autogenerated.  Do not edit */
#define NV_PFAULT                                              /* ----G */
#define NV_PFAULT_MMU_ENG_ID_GRAPHICS          384 /*       */
#define NV_PFAULT_MMU_ENG_ID_DISPLAY           1 /*       */
#define NV_PFAULT_MMU_ENG_ID_GSP               2 /*       */
#define NV_PFAULT_MMU_ENG_ID_IFB               55 /*       */
#define NV_PFAULT_MMU_ENG_ID_FLA               4 /*       */
#define NV_PFAULT_MMU_ENG_ID_BAR1              256 /*       */
#define NV_PFAULT_MMU_ENG_ID_BAR2              320 /*       */
#define NV_PFAULT_MMU_ENG_ID_SEC               6 /*       */
#define NV_PFAULT_MMU_ENG_ID_FSP               7 /*       */
#define NV_PFAULT_MMU_ENG_ID_PERF              10 /*       */
#define NV_PFAULT_MMU_ENG_ID_PERF0             10 /*        */
#define NV_PFAULT_MMU_ENG_ID_PERF1             11 /*        */
#define NV_PFAULT_MMU_ENG_ID_PERF2             12 /*        */
#define NV_PFAULT_MMU_ENG_ID_PERF3             13 /*        */
#define NV_PFAULT_MMU_ENG_ID_PERF4             14 /*        */
#define NV_PFAULT_MMU_ENG_ID_PERF5             15 /*        */
#define NV_PFAULT_MMU_ENG_ID_PERF6             16 /*        */
#define NV_PFAULT_MMU_ENG_ID_PERF7             17 /*        */
#define NV_PFAULT_MMU_ENG_ID_PERF8             18 /*        */
#define NV_PFAULT_MMU_ENG_ID_NVDEC             19 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVDEC0            19 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVDEC1            20 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVDEC2            21 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVDEC3            22 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVDEC4            23 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVDEC5            24 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVDEC6            25 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVDEC7            26 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVJPG0            27 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVJPG1            28 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVJPG2            29 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVJPG3            30 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVJPG4            31 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVJPG5            32 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVJPG6            33 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVJPG7            34 /*       */
#define NV_PFAULT_MMU_ENG_ID_GRCOPY            43 /*       */
#define NV_PFAULT_MMU_ENG_ID_CE0               43 /*       */
#define NV_PFAULT_MMU_ENG_ID_CE1               44 /*       */
#define NV_PFAULT_MMU_ENG_ID_CE2               45 /*       */
#define NV_PFAULT_MMU_ENG_ID_CE3               46 /*       */
#define NV_PFAULT_MMU_ENG_ID_CE4               47 /*       */
#define NV_PFAULT_MMU_ENG_ID_CE5               48 /*       */
#define NV_PFAULT_MMU_ENG_ID_CE6               49 /*       */
#define NV_PFAULT_MMU_ENG_ID_CE7               50 /*       */
#define NV_PFAULT_MMU_ENG_ID_CE8               51 /*       */
#define NV_PFAULT_MMU_ENG_ID_CE9               52 /*       */
#define NV_PFAULT_MMU_ENG_ID_PWR_PMU           5 /*       */
#define NV_PFAULT_MMU_ENG_ID_PTP               3 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVENC0            35 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVENC1            36 /*       */
#define NV_PFAULT_MMU_ENG_ID_NVENC2            37 /*       */
#define NV_PFAULT_MMU_ENG_ID_OFA0              53 /*       */
#define NV_PFAULT_MMU_ENG_ID_PHYSICAL          56 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST0             64 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST1             65 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST2             66 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST3             67 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST4             68 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST5             69 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST6             70 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST7             71 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST8             72 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST9             73 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST10            74 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST11            75 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST12            76 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST13            77 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST14            78 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST15            79 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST16            80 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST17            81 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST18            82 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST19            83 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST20            84 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST21            85 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST22            86 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST23            87 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST24            88 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST25            89 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST26            90 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST27            91 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST28            92 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST29            93 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST30            94 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST31            95 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST32            96 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST33            97 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST34            98 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST35            99 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST36            100 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST37            101 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST38            102 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST39            103 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST40            104 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST41            105 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST42            106 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST43            107 /*       */
#define NV_PFAULT_MMU_ENG_ID_HOST44            108 /*       */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN0          256  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN1          257  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN2          258  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN3          259  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN4          260  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN5          261  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN6          262  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN7          263  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN8          264  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN9          265  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN10         266 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN11         267 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN12         268 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN13         269 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN14         270 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN15         271 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN16         272 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN17         273 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN18         274 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN19         275 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN20         276 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN21         277 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN22         278 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN23         279 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN24         280 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN25         281 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN26         282 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN27         283 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN28         284 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN29         285 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN30         286 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN31         287 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN32         288 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN33         289 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN34         290 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN35         291 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN36         292 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN37         293 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN38         294 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN39         295 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN40         296 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN41         297 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN42         298 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN43         299 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN44         300 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN45         301 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN46         302 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN47         303 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN48         304 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN49         305 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN50         306 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN51         307 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN52         308 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN53         309 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN54         310 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN55         311 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN56         312 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN57         313 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN58         314 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN59         315 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN60         316 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN61         317 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN62         318 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR1_FN63         319 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN0          320  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN1          321  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN2          322  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN3          323  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN4          324  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN5          325  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN6          326  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN7          327  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN8          328  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN9          329  /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN10         330 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN11         331 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN12         332 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN13         333 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN14         334 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN15         335 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN16         336 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN17         337 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN18         338 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN19         339 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN20         340 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN21         341 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN22         342 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN23         343 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN24         344 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN25         345 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN26         346 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN27         347 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN28         348 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN29         349 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN30         350 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN31         351 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN32         352 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN33         353 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN34         354 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN35         355 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN36         356 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN37         357 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN38         358 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN39         359 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN40         360 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN41         361 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN42         362 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN43         363 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN44         364 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN45         365 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN46         366 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN47         367 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN48         368 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN49         369 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN50         370 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN51         371 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN52         372 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN53         373 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN54         374 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN55         375 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN56         376 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN57         377 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN58         378 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN59         379 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN60         380 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN61         381 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN62         382 /*     */
#define NV_PFAULT_MMU_ENG_ID_BAR2_FN63         383 /*     */
#define NV_PFAULT_FAULT_TYPE                             4:0 /*       */
#define NV_PFAULT_FAULT_TYPE_PDE                  0x00000000 /*       */
#define NV_PFAULT_FAULT_TYPE_PDE_SIZE             0x00000001 /*       */
#define NV_PFAULT_FAULT_TYPE_PTE                  0x00000002 /*       */
#define NV_PFAULT_FAULT_TYPE_VA_LIMIT_VIOLATION   0x00000003 /*       */
#define NV_PFAULT_FAULT_TYPE_UNBOUND_INST_BLOCK   0x00000004 /*       */
#define NV_PFAULT_FAULT_TYPE_PRIV_VIOLATION       0x00000005 /*       */
#define NV_PFAULT_FAULT_TYPE_RO_VIOLATION         0x00000006 /*       */
#define NV_PFAULT_FAULT_TYPE_WO_VIOLATION         0x00000007 /*       */
#define NV_PFAULT_FAULT_TYPE_PITCH_MASK_VIOLATION 0x00000008 /*       */
#define NV_PFAULT_FAULT_TYPE_WORK_CREATION        0x00000009 /*       */
#define NV_PFAULT_FAULT_TYPE_UNSUPPORTED_APERTURE 0x0000000a /*       */
#define NV_PFAULT_FAULT_TYPE_COMPRESSION_FAILURE  0x0000000b /*       */
#define NV_PFAULT_FAULT_TYPE_UNSUPPORTED_KIND     0x0000000c /*       */
#define NV_PFAULT_FAULT_TYPE_REGION_VIOLATION     0x0000000d /*       */
#define NV_PFAULT_FAULT_TYPE_POISONED             0x0000000e /*       */
#define NV_PFAULT_FAULT_TYPE_ATOMIC_VIOLATION     0x0000000f /*       */
#define NV_PFAULT_CLIENT                       14:8 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_0        0x00000000 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_1        0x00000001 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_2        0x00000002 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_3        0x00000003 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_4        0x00000004 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_5        0x00000005 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_6        0x00000006 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_7        0x00000007 /*       */
#define NV_PFAULT_CLIENT_GPC_PE_0        0x00000008 /*       */
#define NV_PFAULT_CLIENT_GPC_PE_1        0x00000009 /*       */
#define NV_PFAULT_CLIENT_GPC_PE_2        0x0000000A /*       */
#define NV_PFAULT_CLIENT_GPC_PE_3        0x0000000B /*       */
#define NV_PFAULT_CLIENT_GPC_PE_4        0x0000000C /*       */
#define NV_PFAULT_CLIENT_GPC_PE_5        0x0000000D /*       */
#define NV_PFAULT_CLIENT_GPC_PE_6        0x0000000E /*       */
#define NV_PFAULT_CLIENT_GPC_PE_7        0x0000000F /*       */
#define NV_PFAULT_CLIENT_GPC_RAST        0x00000010 /*       */
#define NV_PFAULT_CLIENT_GPC_GCC         0x00000011 /*       */
#define NV_PFAULT_CLIENT_GPC_GPCCS       0x00000012 /*       */
#define NV_PFAULT_CLIENT_GPC_PROP_0      0x00000013 /*       */
#define NV_PFAULT_CLIENT_GPC_PROP_1      0x00000014 /*       */
#define NV_PFAULT_CLIENT_GPC_PROP_2      0x00000015 /*       */
#define NV_PFAULT_CLIENT_GPC_PROP_3      0x00000016 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_8        0x00000021 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_9        0x00000022 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_10       0x00000023 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_11       0x00000024 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_12       0x00000025 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_13       0x00000026 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_14       0x00000027 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_15       0x00000028 /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_0     0x00000029 /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_1     0x0000002A /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_2     0x0000002B /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_3     0x0000002C /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_4     0x0000002D /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_5     0x0000002E /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_6     0x0000002F /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_7     0x00000030 /*       */
#define NV_PFAULT_CLIENT_GPC_PE_8        0x00000031 /*       */
#define NV_PFAULT_CLIENT_GPC_PE_9        0x00000032 /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_8     0x00000033 /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_9     0x00000034 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_16       0x00000035 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_17       0x00000036 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_18       0x00000037 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_19       0x00000038 /*       */
#define NV_PFAULT_CLIENT_GPC_PE_10       0x00000039 /*       */
#define NV_PFAULT_CLIENT_GPC_PE_11       0x0000003A /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_10    0x0000003B /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_11    0x0000003C /*       */
#define NV_PFAULT_CLIENT_GPC_T1_20       0x0000003D /*       */
#define NV_PFAULT_CLIENT_GPC_T1_21       0x0000003E /*       */
#define NV_PFAULT_CLIENT_GPC_T1_22       0x0000003F /*       */
#define NV_PFAULT_CLIENT_GPC_T1_23       0x00000040 /*       */
#define NV_PFAULT_CLIENT_GPC_PE_12       0x00000041 /*       */
#define NV_PFAULT_CLIENT_GPC_PE_13       0x00000042 /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_12    0x00000043 /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_13    0x00000044 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_24       0x00000045 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_25       0x00000046 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_26       0x00000047 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_27       0x00000048 /*       */
#define NV_PFAULT_CLIENT_GPC_PE_14       0x00000049 /*       */
#define NV_PFAULT_CLIENT_GPC_PE_15       0x0000004A /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_14    0x0000004B /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_15    0x0000004C /*       */
#define NV_PFAULT_CLIENT_GPC_T1_28       0x0000004D /*       */
#define NV_PFAULT_CLIENT_GPC_T1_29       0x0000004E /*       */
#define NV_PFAULT_CLIENT_GPC_T1_30       0x0000004F /*       */
#define NV_PFAULT_CLIENT_GPC_T1_31       0x00000050 /*       */
#define NV_PFAULT_CLIENT_GPC_PE_16       0x00000051 /*       */
#define NV_PFAULT_CLIENT_GPC_PE_17       0x00000052 /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_16    0x00000053 /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_17    0x00000054 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_32       0x00000055 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_33       0x00000056 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_34       0x00000057 /*       */
#define NV_PFAULT_CLIENT_GPC_T1_35       0x00000058 /*       */
#define NV_PFAULT_CLIENT_GPC_PE_18       0x00000059 /*       */
#define NV_PFAULT_CLIENT_GPC_PE_19       0x0000005A /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_18    0x0000005B /*       */
#define NV_PFAULT_CLIENT_GPC_TPCCS_19    0x0000005C /*       */
#define NV_PFAULT_CLIENT_GPC_T1_36       0x0000005D /*       */
#define NV_PFAULT_CLIENT_GPC_T1_37       0x0000005E /*       */
#define NV_PFAULT_CLIENT_GPC_T1_38       0x0000005F /*       */
#define NV_PFAULT_CLIENT_GPC_T1_39       0x00000060 /*       */
#define NV_PFAULT_CLIENT_GPC_ROP_0       0x00000070 /*       */
#define NV_PFAULT_CLIENT_GPC_ROP_1       0x00000071 /*       */
#define NV_PFAULT_CLIENT_GPC_ROP_2       0x00000072 /*       */
#define NV_PFAULT_CLIENT_GPC_ROP_3       0x00000073 /*       */
#define NV_PFAULT_CLIENT_GPC_GPM          0x00000017 /*       */
#define NV_PFAULT_CLIENT_HUB_VIP         0x00000000 /*       */
#define NV_PFAULT_CLIENT_HUB_CE0         0x00000001 /*       */
#define NV_PFAULT_CLIENT_HUB_CE1         0x00000002 /*       */
#define NV_PFAULT_CLIENT_HUB_DNISO       0x00000003 /*       */
#define NV_PFAULT_CLIENT_HUB_DISPNISO    0x00000003 /*       */
#define NV_PFAULT_CLIENT_HUB_FE0         0x00000004 /*       */
#define NV_PFAULT_CLIENT_HUB_FE          0x00000004 /*       */
#define NV_PFAULT_CLIENT_HUB_FECS0       0x00000005 /*       */
#define NV_PFAULT_CLIENT_HUB_FECS        0x00000005 /*       */
#define NV_PFAULT_CLIENT_HUB_HOST        0x00000006 /*       */
#define NV_PFAULT_CLIENT_HUB_HOST_CPU    0x00000007 /*       */
#define NV_PFAULT_CLIENT_HUB_HOST_CPU_NB 0x00000008 /*       */
#define NV_PFAULT_CLIENT_HUB_ISO         0x00000009 /*       */
#define NV_PFAULT_CLIENT_HUB_MMU         0x0000000A /*       */
#define NV_PFAULT_CLIENT_HUB_NVDEC0      0x0000000B /*       */
#define NV_PFAULT_CLIENT_HUB_NVDEC       0x0000000B /*       */
#define NV_PFAULT_CLIENT_HUB_CE3         0x0000000C /*       */
#define NV_PFAULT_CLIENT_HUB_NVENC1      0x0000000D /*       */
#define NV_PFAULT_CLIENT_HUB_NISO        0x0000000E /*       */
#define NV_PFAULT_CLIENT_HUB_ACTRS       0x0000000E /*       */
#define NV_PFAULT_CLIENT_HUB_P2P         0x0000000F /*       */
#define NV_PFAULT_CLIENT_HUB_PD          0x00000010 /*       */
#define NV_PFAULT_CLIENT_HUB_PERF0       0x00000011 /*       */
#define NV_PFAULT_CLIENT_HUB_PERF        0x00000011 /*       */
#define NV_PFAULT_CLIENT_HUB_PMU         0x00000012 /*       */
#define NV_PFAULT_CLIENT_HUB_RASTERTWOD  0x00000013 /*       */
#define NV_PFAULT_CLIENT_HUB_SCC         0x00000014 /*       */
#define NV_PFAULT_CLIENT_HUB_SCC_NB      0x00000015 /*       */
#define NV_PFAULT_CLIENT_HUB_SEC         0x00000016 /*       */
#define NV_PFAULT_CLIENT_HUB_SSYNC       0x00000017 /*       */
#define NV_PFAULT_CLIENT_HUB_GRCOPY      0x00000018 /*       */
#define NV_PFAULT_CLIENT_HUB_CE2         0x00000018 /*       */
#define NV_PFAULT_CLIENT_HUB_XV          0x00000019 /*       */
#define NV_PFAULT_CLIENT_HUB_MMU_NB      0x0000001A /*       */
#define NV_PFAULT_CLIENT_HUB_NVENC0      0x0000001B /*       */
#define NV_PFAULT_CLIENT_HUB_NVENC       0x0000001B /*       */
#define NV_PFAULT_CLIENT_HUB_DFALCON     0x0000001C /*       */
#define NV_PFAULT_CLIENT_HUB_SKED0       0x0000001D /*       */
#define NV_PFAULT_CLIENT_HUB_SKED        0x0000001D /*       */
#define NV_PFAULT_CLIENT_HUB_AFALCON     0x0000001E /*       */
#define NV_PFAULT_CLIENT_HUB_DONT_CARE   0x0000001F /*       */
#define NV_PFAULT_CLIENT_HUB_HSCE0       0x00000020 /*       */
#define NV_PFAULT_CLIENT_HUB_HSCE1       0x00000021 /*       */
#define NV_PFAULT_CLIENT_HUB_HSCE2       0x00000022 /*       */
#define NV_PFAULT_CLIENT_HUB_HSCE3       0x00000023 /*       */
#define NV_PFAULT_CLIENT_HUB_HSCE4       0x00000024 /*       */
#define NV_PFAULT_CLIENT_HUB_HSCE5       0x00000025 /*       */
#define NV_PFAULT_CLIENT_HUB_HSCE6       0x00000026 /*       */
#define NV_PFAULT_CLIENT_HUB_HSCE7       0x00000027 /*       */
#define NV_PFAULT_CLIENT_HUB_HSCE8       0x00000028 /*       */
#define NV_PFAULT_CLIENT_HUB_HSCE9       0x00000029 /*       */
#define NV_PFAULT_CLIENT_HUB_HSHUB       0x0000002A /*       */
#define NV_PFAULT_CLIENT_HUB_PTP_X0      0x0000002B /*       */
#define NV_PFAULT_CLIENT_HUB_PTP_X1      0x0000002C /*       */
#define NV_PFAULT_CLIENT_HUB_PTP_X2      0x0000002D /*       */
#define NV_PFAULT_CLIENT_HUB_PTP_X3      0x0000002E /*       */
#define NV_PFAULT_CLIENT_HUB_PTP_X4      0x0000002F /*       */
#define NV_PFAULT_CLIENT_HUB_PTP_X5      0x00000030 /*       */
#define NV_PFAULT_CLIENT_HUB_PTP_X6      0x00000031 /*       */
#define NV_PFAULT_CLIENT_HUB_PTP_X7      0x00000032 /*       */
#define NV_PFAULT_CLIENT_HUB_NVENC2      0x00000033 /*       */
#define NV_PFAULT_CLIENT_HUB_VPR_SCRUBBER0 0x00000034 /*       */
#define NV_PFAULT_CLIENT_HUB_VPR_SCRUBBER1 0x00000035 /*       */
#define NV_PFAULT_CLIENT_HUB_DWBIF       0x00000036 /*       */
#define NV_PFAULT_CLIENT_HUB_FBFALCON    0x00000037 /*       */
#define NV_PFAULT_CLIENT_HUB_CE_SHIM     0x00000038 /*       */
#define NV_PFAULT_CLIENT_HUB_GSP         0x00000039 /*       */
#define NV_PFAULT_CLIENT_HUB_NVDEC1      0x0000003A /*       */
#define NV_PFAULT_CLIENT_HUB_NVDEC2      0x0000003B /*       */
#define NV_PFAULT_CLIENT_HUB_NVJPG0      0x0000003C /*       */
#define NV_PFAULT_CLIENT_HUB_NVDEC3      0x0000003D /*       */
#define NV_PFAULT_CLIENT_HUB_NVDEC4      0x0000003E /*       */
#define NV_PFAULT_CLIENT_HUB_OFA0        0x0000003F /*       */
#define NV_PFAULT_CLIENT_HUB_HSCE10      0x00000040 /*       */
#define NV_PFAULT_CLIENT_HUB_HSCE11      0x00000041 /*       */
#define NV_PFAULT_CLIENT_HUB_HSCE12      0x00000042 /*       */
#define NV_PFAULT_CLIENT_HUB_HSCE13      0x00000043 /*       */
#define NV_PFAULT_CLIENT_HUB_HSCE14      0x00000044 /*       */
#define NV_PFAULT_CLIENT_HUB_HSCE15      0x00000045 /*       */
#define NV_PFAULT_CLIENT_HUB_PTP_X8      0x00000046 /*       */
#define NV_PFAULT_CLIENT_HUB_PTP_X9      0x00000047 /*       */
#define NV_PFAULT_CLIENT_HUB_PTP_X10     0x00000048 /*       */
#define NV_PFAULT_CLIENT_HUB_PTP_X11     0x00000049 /*       */
#define NV_PFAULT_CLIENT_HUB_PTP_X12     0x0000004A /*       */
#define NV_PFAULT_CLIENT_HUB_PTP_X13     0x0000004B /*       */
#define NV_PFAULT_CLIENT_HUB_PTP_X14     0x0000004C /*       */
#define NV_PFAULT_CLIENT_HUB_PTP_X15     0x0000004D /*       */
#define NV_PFAULT_CLIENT_HUB_FE1         0x0000004E /*       */
#define NV_PFAULT_CLIENT_HUB_FE2         0x0000004F /*       */
#define NV_PFAULT_CLIENT_HUB_FE3         0x00000050 /*       */
#define NV_PFAULT_CLIENT_HUB_FE4         0x00000051 /*       */
#define NV_PFAULT_CLIENT_HUB_FE5         0x00000052 /*       */
#define NV_PFAULT_CLIENT_HUB_FE6         0x00000053 /*       */
#define NV_PFAULT_CLIENT_HUB_FE7         0x00000054 /*       */
#define NV_PFAULT_CLIENT_HUB_FECS1       0x00000055 /*       */
#define NV_PFAULT_CLIENT_HUB_FECS2       0x00000056 /*       */
#define NV_PFAULT_CLIENT_HUB_FECS3       0x00000057 /*       */
#define NV_PFAULT_CLIENT_HUB_FECS4       0x00000058 /*       */
#define NV_PFAULT_CLIENT_HUB_FECS5       0x00000059 /*       */
#define NV_PFAULT_CLIENT_HUB_FECS6       0x0000005A /*       */
#define NV_PFAULT_CLIENT_HUB_FECS7       0x0000005B /*       */
#define NV_PFAULT_CLIENT_HUB_SKED1       0x0000005C /*       */
#define NV_PFAULT_CLIENT_HUB_SKED2       0x0000005D /*       */
#define NV_PFAULT_CLIENT_HUB_SKED3       0x0000005E /*       */
#define NV_PFAULT_CLIENT_HUB_SKED4       0x0000005F /*       */
#define NV_PFAULT_CLIENT_HUB_SKED5       0x00000060 /*       */
#define NV_PFAULT_CLIENT_HUB_SKED6       0x00000061 /*       */
#define NV_PFAULT_CLIENT_HUB_SKED7       0x00000062 /*       */
#define NV_PFAULT_CLIENT_HUB_ESC         0x00000063 /*       */
#define NV_PFAULT_CLIENT_HUB_NVDEC5      0x0000006F /*       */
#define NV_PFAULT_CLIENT_HUB_NVDEC6      0x00000070 /*       */
#define NV_PFAULT_CLIENT_HUB_NVDEC7      0x00000071 /*       */
#define NV_PFAULT_CLIENT_HUB_NVJPG1      0x00000072 /*       */
#define NV_PFAULT_CLIENT_HUB_NVJPG2      0x00000073 /*       */
#define NV_PFAULT_CLIENT_HUB_NVJPG3      0x00000074 /*       */
#define NV_PFAULT_CLIENT_HUB_NVJPG4      0x00000075 /*       */
#define NV_PFAULT_CLIENT_HUB_NVJPG5      0x00000076 /*       */
#define NV_PFAULT_CLIENT_HUB_NVJPG6      0x00000077 /*       */
#define NV_PFAULT_CLIENT_HUB_NVJPG7      0x00000078 /*       */
#define NV_PFAULT_CLIENT_HUB_FSP         0x00000079 /*       */
#define NV_PFAULT_ACCESS_TYPE                 19:16 /*       */
#define NV_PFAULT_ACCESS_TYPE_READ       0x00000000 /*       */
#define NV_PFAULT_ACCESS_TYPE_WRITE      0x00000001 /*       */
#define NV_PFAULT_ACCESS_TYPE_ATOMIC     0x00000002 /*       */
#define NV_PFAULT_ACCESS_TYPE_PREFETCH   0x00000003 /*       */
#define NV_PFAULT_ACCESS_TYPE_VIRT_READ          0x00000000 /*       */
#define NV_PFAULT_ACCESS_TYPE_VIRT_WRITE         0x00000001 /*       */
#define NV_PFAULT_ACCESS_TYPE_VIRT_ATOMIC        0x00000002 /*       */
#define NV_PFAULT_ACCESS_TYPE_VIRT_ATOMIC_STRONG 0x00000002 /*       */
#define NV_PFAULT_ACCESS_TYPE_VIRT_PREFETCH      0x00000003 /*       */
#define NV_PFAULT_ACCESS_TYPE_VIRT_ATOMIC_WEAK   0x00000004 /*       */
#define NV_PFAULT_ACCESS_TYPE_PHYS_READ          0x00000008 /*       */
#define NV_PFAULT_ACCESS_TYPE_PHYS_WRITE         0x00000009 /*       */
#define NV_PFAULT_ACCESS_TYPE_PHYS_ATOMIC        0x0000000a /*       */
#define NV_PFAULT_ACCESS_TYPE_PHYS_PREFETCH      0x0000000b /*       */
#define NV_PFAULT_MMU_CLIENT_TYPE             20:20 /*       */
#define NV_PFAULT_MMU_CLIENT_TYPE_GPC    0x00000000 /*       */
#define NV_PFAULT_MMU_CLIENT_TYPE_HUB    0x00000001 /*       */
#define NV_PFAULT_GPC_ID                      28:24 /*       */
#define NV_PFAULT_PROTECTED_MODE              29:29 /*       */
#define NV_PFAULT_REPLAYABLE_FAULT_EN         30:30 /*       */
#define NV_PFAULT_VALID                       31:31 /*       */
#endif // __gh100_dev_fault_h__
